/* alocacao.c */

/* Funcoes reponsaveis pelas alocacoes dos nodos e manipulacao de excecao */

#include <alloc.h>
#include <conio.h>
#include "puzzle.h"
#include "alocacao.h"

#ifdef PC
#include <graphics.h>
#endif

extern word nos_gerados;
extern no inicial;
extern void interrupt far (*velho_clock_int)(void);

/* aloca_no: Aloca espaco no heap e encadeia a lista de nos alocados para
   uma futura proximo_no.
   Recebe: Nada.
   Retorna: O endereco do no na memoria. */

no aloca_no (void)
{
  static no ultimo_no = NULL;
  no ptr_novo_no;

  ptr_novo_no = farmalloc (sizeof (struct no));
  if (ptr_novo_no == NULL)
    falta_memoria ();

  nos_gerados++;

  /* Encadeia o no na lista do ja alocados */
  if (ultimo_no == NULL)
    ultimo_no = ptr_novo_no;
  else {
    ultimo_no->proximo_no = ptr_novo_no;
    ultimo_no = ptr_novo_no;
  }
  ptr_novo_no->proximo_no = NULL;

  return ptr_novo_no;
}

/* desaloca_nos: Desencadeia o procedimento de desalocacao dos no na memoria
   heap. Isto e feito seguindo a lista dos nos alocado, encabecado pelo no
   inicial.
   Recebe: O no inicial (primeiro no aloca_no).
   Retorna: Nada. */

void desaloca_nos (no inicial)
{
  no atual,
     posterior;

  atual = inicial;
  do {
    posterior = atual->proximo_no;
    farfree ((void far *) atual);
    atual = posterior;
  } while (posterior != NULL);
}

void aloca_sucessores (no sucessores [])
{
   byte cont;

   for (cont = 0; cont < MAXSUCESSORES; cont++)
     sucessores [cont] = aloca_no ();
}

/* aloca_sucessor: Constroi uma replica do nodo sucessor e retorna o seu
   ponteiro.
   Recebe: Um ponteiro para o no origem;
   Retorna: Um ponteiro para a sua copia na memoria. */

no aloca_sucessor (no origem)
{
  no sucessor_alocado;

  sucessor_alocado = aloca_no ();
  inicialize (sucessor_alocado);

  /* Copia tabuleiro */
  sucessor_alocado->tabuleiro = origem->tabuleiro;
  sucessor_alocado->g = origem->g;
  sucessor_alocado->h = origem->h;
  sucessor_alocado->melhor_no = origem->melhor_no;
  sucessor_alocado->proximo_fila = sucessor_alocado->proximo_fila;
  return sucessor_alocado;
}

/* falta_memoria: Termina o programa informando que nao ha memoria
   suficiente para alocacao de nos.
   Recebe: Nada.
   Retorna: Nada. */

void falta_memoria (void)
{

  desaloca_nos (inicial);

#ifdef PC
  finaliza_estatistica_PC ();
  closegraph ();
#endif

  clrscr ();
  printf ("Puzzle: Faltou memoria para a alocacao do no.\n");
  printf ("Verifique se a solucao e realmente possivel.\n");
  exit (0);

}