/* interfac.c */

/* Implementacao da interface com o usuario, tanto para equipamento IBM PC
   quanto para outros equipamentos (usando somente stdio). */

#include <stdio.h>
#include <alloc.h>
#include "puzzle.h"
#include "interfac.h"

#ifdef PC
#include <conio.h>
#include <dos.h>
#include <graphics.h>
#include <bios.h>
#include <alloc.h>
#include <stdlib.h>
#endif

word nivel_arvore = 0, /* Nivel maximo da arvore ate o momento */
     nos_gerados = 0,  /* Indica o numero toral de nodos gerados */
     nos_tratados = 0; /* Indica o numero de nos tratados (fechados) */

#ifdef PC

byte linha = 1,     /* Posicoes do cursor grafico no modo PC */
     coluna = 1;

char ch_nivel_arvore [5],
     ch_nos_abertos [8],
     ch_nos_gerados [8],
     ch_nos_tratados [8],
     ch_total_nos [8],
     ch_memoria [10];

void interrupt far (*velho_clock_int)(void);

/* cursor: Produz um cursor intermitente em modo grafico.
   Recebe: Nada.
   Retorna: Nada. */

void interrupt far cursor (void)
{
  static int linha_ant = 1,
	     coluna_ant = 1,
	     passagem = 0;
  int cor, pos_x, pos_y;

  if (linha < 1)
    linha = 1;
  if (linha > 3)
    linha = 3;
  if (coluna < 1)
    coluna = 1;
  if (coluna > 3)
    coluna = 3;
  if (passagem++ == 3) {
    cor = getcolor ();
    pos_x = 120 + (coluna - 1) * 40 - 4;
    pos_y = 60  + (linha - 1) * 30 - 4;
    setcolor (VERMELHO);
    rectangle (pos_x, pos_y, pos_x + 34, pos_y + 24);
    setcolor (cor);
    linha_ant = linha;
    coluna_ant = coluna;
    return;
  }
  if (passagem == 6) {
    cor = getcolor ();
    pos_x = 120 + (coluna_ant - 1) * 40 - 4;
    pos_y = 60  + (linha_ant - 1) * 30 - 4;
    setcolor (PRETO);
    rectangle (pos_x, pos_y, pos_x + 34, pos_y + 24);
    setcolor (cor);
    passagem = 0;
    return;
  }
}

/* recebe_no_inicial: Recebe a configuracao inicial do tabuleiro do Puzzle.
   Recebe: Um ponterio para o no que recebera as informacoes.
   Retorna: Nada. */

void recebe_no_inicial (no ptr_no)
{
  byte valor;
  int varredura, tecla;
  int device_grf, modo_grf;

  device_grf = CGA;
  modo_grf = CGAC0;
  initgraph (&device_grf, &modo_grf, ".");

  settextstyle (DEFAULT_FONT, HORIZ_DIR, 3);
  setcolor (VERDE);
  outtextxy (5, 15, "P");
  outtextxy (5, 45, "U");
  outtextxy (5, 75, "Z");
  outtextxy (5, 105, "Z");
  outtextxy (5, 135, "L");
  outtextxy (5, 165, "E");

  setlinestyle (DASHED_LINE, EMPTY_FILL, 1);
  setcolor (VERMELHO) ;
  rectangle (40, 5, 310, 195);

  setlinestyle (SOLID_LINE, SOLID_FILL, 1);
  setcolor (AMARELO);
  rectangle (110, 50, 240, 150);

  settextstyle (SMALL_FONT, HORIZ_DIR, 0);
  imprima (50, 160, "Use as setas para por as peas");
  imprima (50, 170, "no tabuleiro. Use <Enter> para");
  imprima (50, 180, "iniciar a busca da soluo.   ");
  velho_clock_int = getvect (CLOCK);
  setvect (CLOCK, cursor);
  do {
    int verredura;
    do {
      while ((varredura = bioskey (1)) == 0)
      ;
      tecla = getch ();
      switch (varredura) {
	case SOBE:
	  if (--linha < 1)
	    linha = 1;
	  getch ();
	  break;
	case DESCE:
	  if (++linha > 3)
	    linha = 3;
	  getch ();
	  break;
	case DIREITA:
	  if (++coluna > 3)
	    coluna = 3;
	  getch ();
	  break;
	case ESQUERDA:
	  if (--coluna < 1)
	    coluna = 1;
	  getch ();
	  break;
	default: {
	  valor = ORD (tecla);
	  if (tecla != ENTER)
	    if (valor >= 1 && valor <= 8) {
	      retira_peca ();
	      setar_posicao (ptr_no, linha, coluna, valor);
	      forma_peca (valor);
	    }
	    else {
	      setar_posicao (ptr_no, linha, coluna, 0);
	      retira_peca ();
	    }
        }
      }
    } while (tecla != ENTER);
  } while (!configuracao_valida (ptr_no));
  setvect (CLOCK, velho_clock_int);
  limpa_mensagens ();
}

/* retira_cursor: Limpa da tela, os tracos deixados pelo cursor.
   Recebe: Nada.
   Retorna: Nada. */

void limpa_mensagens (void)
{
  byte linha,
       coluna,
       pos_x,
       pos_y,
       cor;

  cor = getcolor ();
  setcolor (PRETO);
  for (linha = 0; linha < 3; linha++)
    for (coluna = 0; coluna < 3; coluna++) {
      pos_x = 120 + coluna * 40 - 4;
      pos_y = 60  + linha * 30 - 4;
      rectangle (pos_x, pos_y, pos_x + 34, pos_y + 24);
    }
  setfillstyle (EMPTY_FILL, PRETO);
  bar (43, 153, 307, 192);
  setcolor (cor);
}

/* forma_peca: Desenha no tabuleiro a peca estipulada.
   Recebe: O valor da peca.
   Retorna: Nada. */

void forma_peca (byte valor)
{
  int pos_x, pos_y;
  char letra [2];

  pos_x = 120 + (coluna - 1) * 40 - 2;
  pos_y = 60  + (linha - 1) * 30 - 2;
  setfillstyle (LTBKSLASH_FILL, AMARELO);
  bar (pos_x, pos_y, pos_x + 30, pos_y + 20);
  setcolor (VERDE);
  line (pos_x + 30, pos_y + 2, pos_x + 30, pos_y + 20);
  line (pos_x + 2, pos_y + 20, pos_x + 30, pos_y + 20);
  setfillstyle (SOLID_FILL, AMARELO);
  bar (pos_x + 2, pos_y + 2, pos_x + 10, pos_y + 10);
  setcolor (VERMELHO);
  letra [1] = '\0';
  letra [0] = CHR (valor);
  outtextxy (pos_x + 3, pos_y + 3, letra);
}

/* retira_peca: Apaga uma peca do tabuleiro.
   Recebe: Nada.
   Retorna: Nada. */

void retira_peca (void)
{
  int pos_x, pos_y;

  pos_x = 120 + (coluna - 1) * 40 - 2;
  pos_y = 60  + (linha - 1) * 30 - 2;
  setfillstyle (EMPTY_FILL, PRETO);
  bar (pos_x, pos_y, pos_x + 30, pos_y + 20);
}

/* configuracao_valida: Diz se as pecas do tabuleiro sao validas.
   Recebe: O nodo a ser analisado.
   Returna: TRUE se o tabuleiro for valido ou FALSE caso contrario. */

byte configuracao_valida (no ptr_no)
{
  byte linha,
       coluna,
       peca = 0,
       estado;

  do {
    estado = FALSE;
    for (linha = 1; linha <= 3; linha++)
      for (coluna = 1; coluna <= 3; coluna++)
	if (posicao (TABULEIRO (ptr_no), linha, coluna) == peca) {
	  peca++;
	  estado = TRUE;
	}

  } while (estado == TRUE && peca != 9);
  return (estado == TRUE ? TRUE : FALSE);
}

/* imprima: Mostra uma mensagem sequencializando as letra por tempo.
   Recebe: As coordenadas de linha, coluna e a mensagem.
   Retorna: Nada. */

void imprima (int x, int y, char *mensagem)
{
  int cont = 0;
  char letra [2];

  letra [1] = '\0';
  moveto (x, y);
  for (; mensagem [cont] != '\0'; cont++) {
    delay (50);
    letra [0] = mensagem [cont];
    outtext (letra);
  }
}

/* mostra_tabuleiro: Reproduz o tabuleiro do nodo no video.
   Recebe: O nodo a ser mostrado.
   Retorna: Nada. */

void mostra_tabuleiro (no ptr_no)
{
  byte linha,
       coluna,
       linha_vazio_ant,
       coluna_vazio_ant,
       linha_vazio,
       coluna_vazio,
       x1, y1, x2, y2,
       cont;
  int incr;
  static long solucao_ant = 0l;
  long tam;
  void far *peca;

  if (solucao_ant == 0l) {
    solucao_ant = TABULEIRO (ptr_no);
    return;
  }

  for (linha = 1; linha <= 3; linha++)
    for (coluna = 1; coluna <= 3; coluna++) {
      if (posicao (TABULEIRO (ptr_no), linha, coluna) == 0) {
	linha_vazio = linha;
	coluna_vazio = coluna;
      }
      if (posicao (solucao_ant, linha, coluna) == 0) {
	linha_vazio_ant = linha;
	coluna_vazio_ant = coluna;
      }
    }
  getch ();
  x1 = 120 + (coluna_vazio - 1) * 40;
  y1 = 60  + (linha_vazio - 1) * 30;
  x2 = x1 + 30;
  y2 = y1 + 20;
  tam = imagesize (x1 - 4, y1 - 4, x2 + 4, y2 + 4);
  peca = farmalloc (tam);
  getimage (x1 - 4, y1 - 4, x2 + 4, y2 + 4, peca);
  if (coluna_vazio_ant != coluna_vazio) {
    if (coluna_vazio_ant < coluna_vazio)
      incr = -1;
    else
      incr = 1;
    for (cont = 0; cont < 41; cont++, x1 += incr)
      putimage (x1 - 4, y1 - 4, peca, 0);
  }
  else {
    if (linha_vazio_ant < linha_vazio)
      incr = -1;
    else
      incr = 1;
    for (cont = 0; cont < 31; cont++, y1 += incr)
      putimage (x1 - 4, y1 - 4, peca, 0);
  }
  farfree (peca);
  solucao_ant = TABULEIRO (ptr_no);
}

/* mostra_caminho: Mostrando o caminho ate o nodo meta.
   Recebe: O nodo meta.
   Retorna: Nada. */

void mostra_caminho (no meta)
{
  no auxiliar,
     problema;

  auxiliar = meta;
  while (auxiliar->melhor_no != NULL)
    auxiliar = auxiliar->melhor_no;
  problema = auxiliar;
  mostra_tabuleiro (problema);

  setcolor (AMARELO);
  outtextxy (47, 100, "Soluo");
  outtextxy (245, 100, "<Enter>");
  setcolor (VERMELHO);
  line (47, 110, 102, 110);
  line (245, 110, 300, 110);
  do {
    auxiliar = meta;
    if (auxiliar->melhor_no != NULL) {
      while (!TABULEIROS_IGUAIS (auxiliar->melhor_no, problema))
	auxiliar = auxiliar->melhor_no;
      mostra_tabuleiro (auxiliar);
    }
    problema = auxiliar;
  } while (!TABULEIROS_IGUAIS (meta, problema));
  getch ();
  closegraph ();
  cprintf ("Fim Puzzle\n\r");
}

/* estatistica: Fornece informacoes a respeito do estado da procura pelo no
   meta.
   Recebe: Nada.
   Retorna: Nada. */

void interrupt far estatistica_PC (void)
{
  static int primeiro = 0;

  if (primeiro == 0) {
    setcolor (AMARELO);
    outtextxy (50, 15, "Ns abertos..:");
    outtextxy (50, 25, "Ns tratados.:");
    outtextxy (50, 35, "Total ns....:");
    outtextxy (50, 160, "Memria......:");
    outtextxy (50, 175, "Nivel rvore.:");
    primeiro = 1;
  }
  if (++primeiro != 20)
    return;
  primeiro = 1;
  setfillstyle (EMPTY_FILL, PRETO);
  bar (175, 13, 230, 43);
  bar (180, 157, 250, 185);
  setcolor (VERMELHO);
  outtextxy (200, 160, ch_memoria);
  outtextxy (200, 15, ch_nos_abertos);
  outtextxy (200, 25, ch_nos_tratados);
  outtextxy (200, 35, ch_nos_gerados);
  outtextxy (200, 175, ch_nivel_arvore);
}

/* inicializa_estatistica_PC: Prepara a interrupcoes para a estatistica.
   Recebe: Nada.
   Retorna: Nada. */

void inicializa_estatistica_PC ()
{
  velho_clock_int = getvect (CLOCK);
  setvect (CLOCK, estatistica_PC);
}

/* finaliza_estatistica_PC: Restaura as interrupcoes para da estatistica.
   Recebe: Nada.
   Retorna: Nada. */

void finaliza_estatistica_PC ()
{
  setvect (CLOCK, velho_clock_int);
}

#else

/* recebe_no_inicial: Recebe, atraves do stdin, a configuracao inicial do
   tabuleiro do Puzzle.
   Recebe: Um ponterio para o no que recebera as informacoes.
   Retorna: Nada. */

void recebe_no_inicial (no ptr_no)
{
  byte linha,
       coluna,
       valor;

  printf ("\nPuzzle\n\nForneca a configuracao inicial:\n");
  for (linha = 1; linha <= 3; linha++)
    for (coluna = 1; coluna <= 3; coluna++) {
      printf ("\nLinha %d coluna %d valor: ", linha, coluna);
      valor = ORD (getc (stdin));
      getc (stdin); /* Retira o <Enter> do buffer */
      setar_posicao (ptr_no, linha, coluna, valor);
    }
  printf ("\n\n");
}

/* mostra_tabuleiro: Reproduz o tabuleiro do nodo no video.
   Recebe: O nodo a ser mostrado.
   Retorna: Nada. */

void mostra_tabuleiro (no ptr_no)
{
  byte linha,
       coluna;

  printf ("+---+---+---+\n");
  for (linha = 1; linha <= 3; linha++) {
    for (coluna = 1; coluna <= 3; coluna++) {
      printf ("| %d ", posicao (TABULEIRO (ptr_no), linha, coluna));
    }
    printf ("|\n+---+---+---+\n");
  }
  getc (stdin);
  printf ("\n");
}

/* mostra_caminho: Mostrando o caminho ate o nodo meta.
   Recebe: O nodo meta.
   Retorna: Nada. */

void mostra_caminho (no meta)
{
  no auxiliar,
     problema;

  printf ("\n\nSolucao do problema.\n\n");
  auxiliar = meta;
  while (auxiliar->melhor_no != NULL)
    auxiliar = auxiliar->melhor_no;
  problema = auxiliar;
  mostra_tabuleiro (problema);

  do {
    auxiliar = meta;
    if (auxiliar->melhor_no != NULL) {
      while (!TABULEIROS_IGUAIS (auxiliar->melhor_no, problema))
	auxiliar = auxiliar->melhor_no;
      mostra_tabuleiro (auxiliar);
    }
    problema = auxiliar;
  } while (!TABULEIROS_IGUAIS (meta, problema));
  printf ("Fim Puzzle.\n");
}

#endif

/* estatistica: Fornece informacoes a respeito do estado da procura pelo no
   meta.
   Recebe: Nada.
   Retorna: Nada. */

void estatistica (void)
{

#ifdef PC

  ltoa (farcoreleft (), ch_memoria, 10);
  itoa (nos_gerados - nos_tratados, ch_nos_abertos, 10);
  itoa (nos_tratados, ch_nos_tratados, 10);
  itoa (nos_gerados, ch_nos_gerados, 10);
  itoa (nivel_arvore, ch_nivel_arvore, 10);

#else

  printf ("\rMemoria: %li, abertos, %d, tratados %d, total nos %d, arvore %d ",
    farcoreleft (), nos_gerados - nos_tratados, nos_tratados,
    nos_gerados, nivel_arvore);

#endif

}
