/* Puzzle.c */

#include <stdio.h>
#include <mem.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "puzzle.h"
#include "alocacao.h"
#include "interfac.h"
#include "filas.h"

extern word nivel_arvore,
	    nos_tratados;

no inicial;

/* calcule_h: Conta o numero de pecas que nao estao no seu lugar correto no
   tabuleiro. Se for definido 'DISTANCIA' ele calculara a soma das distancias
   das pecas em relacao as suas casas.
   Recebe: O no que sera calculado.
   Retorna: O nivel de erro do tabuleiro. */

word calcule_h (no ptr_no)
{
  byte linha,
       coluna,
       valor_correto = 8, /* Indica que peca deveria estar na linha e coluna */
       nivel_erro = 0;

#ifdef DISTANCIA /* Distancia das pecas de suas casas */

  for (linha = 1; linha <= 3; linha++)
    for (coluna = 1; coluna <= 3; coluna++) {
     byte peca,
	  linha_real,
	  coluna_real,
	  distancia = 0;

     peca = posicao (TABULEIRO (ptr_no), linha, coluna);
     coluna_real = 4 - (peca % 3 + 1); /* Coluna que a peca deveria estar */
     linha_real = 4 - (peca / 3 + 1);  /* Linha que a peca deveria estar */
     distancia += abs ((int) (linha - linha_real)) +
		  abs ((int) (coluna - coluna_real));
     nivel_erro += distancia;
   }

#else /* Pecas em casa erradas */

  for (linha = 1; linha <= 3; linha++)
    for (coluna = 1; coluna <= 3; coluna++)
     if (ptr_no->tabuleiro [POSICAO (linha, coluna)] != valor_correto--)
       nivel_erro++;

#endif

  return nivel_erro;
}

/* posicao: Fornece o valor da peca que esta no tabuleiro nas posicoes
   fornecidas.
   Recebe: O tabuleiro, e as coordenadas da peca desejada. Canto superior
   esquerdo e a cordenada (linha 1, coluna 1).
   Retorna: O valor da peca. */

byte posicao (long tabuleiro, byte linha, byte coluna)
{
  char asc_tabuleiro [10];
  byte valor;

  ltoa (tabuleiro, asc_tabuleiro, 10);
  valor = ORD (asc_tabuleiro [(linha - 1) * 3 + coluna - 1]);
  return (valor == 9? 0 : valor);
}

/* setar_posicao: Coloca uma peca no tabuleiro do nodo fornecido.
   Recebe: O nodo, as coordenadas e o valor da peca.
   Retorna: Nada. */

void setar_posicao (no ptr_no, byte linha, byte coluna, byte valor)
{
  char tabuleiro [10];

  ltoa (ptr_no->tabuleiro, tabuleiro, 10);
  if (valor == 0) /* Casa vazia corresponde ao numero 9 */
    valor = 9;
  tabuleiro [(linha - 1) * 3 + coluna - 1] = CHR (valor);
  ptr_no->tabuleiro = atol (tabuleiro);
}

/* zera_sucessores: Inicializa os sucessores gerados, mas nao tradados.
   Recebe: O ponterio para o vetor (por referencia).
   Retorna: Nada. */

void zera_sucessores (no sucessores [])
{
  byte cont;

  for (cont = 0; cont < MAXSUCESSORES; cont++)
    inicialize (sucessores [cont]);
}

/* inicialize: Retira qualquer lixo existente no nodo, devido a alocacao de
   memoria.
   Recebe: Um ponteiro para o no.
   Retorna: Nada. */

void inicialize (no ptr_no)
{
  byte linha,
       coluna,
       cont;

  ptr_no->tabuleiro = 999999999l; /* Zera tabuleiro */
  ptr_no->g = ptr_no->h = 0l;
  ptr_no->melhor_no = ptr_no->proximo_fila = NULL;
  for (cont = 0; cont < MAXSUCESSORES; cont++)
    ptr_no->sucessores [cont] = NULL;
}

/* acha_melhor_no: Procura o nodo de menor custo na lista dos nodos abertos,
   retira este no da lista e retorna o seu endereco.
   Recebe: O endereco do ponteiro da lista dos abertos. Ele tem que ser
   passados por referencia pois ao retira o melhor no, ele pode ser modificado.
   Retorna: O ponteiro para o melhor no. */

no acha_melhor_no (no *aberto)
{
  no melhor_ant,      /* Nodo anterior ao melhor no. Isto para possi-
			 bilitar a retirada do melhor no da lista. */
     melhor,          /* Melhor no encontrado na busca */
     atual,
     proximo;

  byte primeiro_nodo = TRUE; /* A retirada do primeiro nodo da lista requer
				tratamento especial. */

  if (*aberto == NULL)
    return NULL;
  atual = melhor = *aberto;       /* Primeiro no */
  proximo = atual->proximo_fila;  /* Segundo no */
  if (proximo == NULL) {
    *aberto = NULL;
    return melhor;
  }
  else {
    while (proximo != NULL) {
      if (F (melhor) > F(proximo)) {
	melhor = proximo;
	melhor_ant = atual;
	primeiro_nodo = FALSE;
      }
      atual = proximo;
      proximo = proximo->proximo_fila;
    }
    if (primeiro_nodo)
      *aberto = melhor->proximo_fila;
    else
       melhor_ant->proximo_fila = melhor->proximo_fila;
    melhor->proximo_fila = NULL;
    return melhor;
  }
}

/*  cria_sucessor: Gera o sucessor com o movimento da peca especificado.
    Recebe: A linha e coluna da peca que sera movida, a posicao do tabuleiro
    que esta vazia e o no pai.
    Retorna: O no sucessor com todos os dados preparados. */

void cria_sucessor (byte linha_ant, byte coluna_ant,
		   byte linha_atu, byte coluna_atu, no ptr_no, no destino)
{

  /* Copia tabuleiro */
  destino->tabuleiro = ptr_no->tabuleiro;

  /* Executa movimento */
  setar_posicao (destino, linha_atu, coluna_atu,
    posicao (TABULEIRO (ptr_no), linha_ant, coluna_ant));
  setar_posicao (destino, linha_ant, coluna_ant, 0);
}

/* gera_sucessores: Produz todos os sucessores possiveis de um dado no.
   Recebe: O endereco de um vetor de ponteiros para sucessores e um ponteiro
   para o nodo pai.
   Retorna: Os sucessores (passagem por referencia) do nodo dado. */

byte gera_sucessores (no sucessores [], no ptr_no)
{
  byte linha, coluna,
       vazio_linha, vazio_coluna; /* Posicao da casa vazia no tabuleiro */
  byte raiz,
       gerado = 0;

  /* Acha a casa vazia do tabuleiro */
  for (linha = 1; linha <= 3; linha++)
    for (coluna = 1; coluna <= 3; coluna++)
      if (posicao (TABULEIRO (ptr_no), linha, coluna) == 0) {
	vazio_linha = linha;
	vazio_coluna = coluna;
      }
  if (ptr_no->melhor_no == NULL)
    raiz = TRUE;
  else
    raiz = FALSE;
  /* Gera os sucessore evitando a repeticao do nodo pai */
  if (vazio_linha < 3) { /* Pode-se mover de baixo para cima */
    if (raiz || posicao (TABULEIRO (ptr_no), vazio_linha + 1, vazio_coluna) !=
	   posicao (TABULEIRO (ptr_no->melhor_no), vazio_linha, vazio_coluna))
      cria_sucessor (vazio_linha + 1, vazio_coluna,
	   vazio_linha, vazio_coluna, ptr_no, sucessores [gerado++]);
  }
  if (vazio_linha > 1) { /* Pode-se mover de cima para baixo */
    if (raiz || posicao (TABULEIRO (ptr_no), vazio_linha - 1, vazio_coluna) !=
	   posicao (TABULEIRO (ptr_no->melhor_no), vazio_linha, vazio_coluna))
      cria_sucessor (vazio_linha - 1, vazio_coluna,
	   vazio_linha, vazio_coluna, ptr_no, sucessores [gerado++]);
  }
  if (vazio_coluna < 3) { /* Pode-se mover da direita para esquerda */
    if (raiz || posicao (TABULEIRO (ptr_no), vazio_linha, vazio_coluna + 1) !=
	   posicao (TABULEIRO (ptr_no->melhor_no), vazio_linha, vazio_coluna))
      cria_sucessor (vazio_linha, vazio_coluna + 1,
	   vazio_linha, vazio_coluna, ptr_no, sucessores [gerado++]);
  }
  if (vazio_coluna > 1) { /* Pode-se mover da esquerda para direita */
    if (raiz || posicao (TABULEIRO (ptr_no), vazio_linha, vazio_coluna - 1) !=
	   posicao (TABULEIRO (ptr_no->melhor_no), vazio_linha, vazio_coluna))
      cria_sucessor (vazio_linha, vazio_coluna - 1,
	   vazio_linha, vazio_coluna, ptr_no, sucessores [gerado++]);
  }
  return gerado;
}

/* atualize_nivel_arvore: Indicativo do nivel da arvore gerado ate o momento.
   Recebe: O melhor no encontrado na lista dos melhores nodos.
   Retorna: Nada */

void atualize_nivel_arvore (no ptr_no)
{
  if (G (ptr_no) > nivel_arvore)
    nivel_arvore = G (ptr_no);
}

/* esta_em: Verifica se existe um no na fila cujo tabuleiro e igual ao nodo
   passado.
   Recebe: Um ponteiro para a fila, outro para o no e um ponteiro passado por
   referencia para retornar o no achado caso ele exista.
   Retorna: Verdadeiro de o no possui outro igual na fila, falso caso
   contrario, um ponteiro para o no na fila, caso ele exista. */

byte esta_em (no fila, no ptr_no, no *antigo)
{
  no nodo_fila;

  nodo_fila = fila;
  *antigo = NULL;
  while (nodo_fila != NULL)
    if (TABULEIROS_IGUAIS (nodo_fila, ptr_no)) {
      *antigo = nodo_fila;
      return TRUE;
    }
    else
      nodo_fila = nodo_fila->proximo_fila;
  return FALSE;
}

/* acrescenta_sucessor: Coloca o no recebido na lista dos sucessores.
   Recebe: Um ponteiro para o nodo pai e outro para o nodo filho.
   Retorna: Nada. */

void acrescenta_sucessor (no pai, no filho)
{
  byte posicao = 0;

  while (pai->sucessores [posicao] != NULL)
    if (posicao == MAXSUCESSORES - 1) {
      printf ("Puzzle: Erro fatal, numero excessivo de sucessores.\n");
      exit (0);
    }
  else
    posicao++;
  pai->sucessores [posicao] = filho;
}

/* expanda_f: Propaga o valor de g e f para os nodos filhos, netos ...
   Recebe: O nodo pai e o seu g.
   Retorna: Nada. */

void expanda_f (no nodo, byte g)
{
  byte cont = 0;

  while (nodo->sucessores [cont] != NULL && cont < MAXSUCESSORES) {
    if (nodo->sucessores [cont]->melhor_no == nodo) {
      expanda_f (nodo->sucessores [cont], G (nodo) + 1);
      FIXE_G (nodo, g);
    }
    else
      if (g < G (nodo->sucessores [cont])) {
	expanda_f (nodo->sucessores [cont], G (nodo) + 1);
	FIXE_G (nodo, g);
      }
    cont++;
  }
}

/* Implementacao do algoritmo A* para o problema do puzzle de oito pecas. */

main ()
{
  no aberto = NULL,
     fechado = NULL,
     melhor_no = NULL,
     antigo = NULL,
     sucessores [MAXSUCESSORES],
     sucessor_alocado;

  no no_meta = NULL;

  inicial = aloca_no ();
  inicialize (inicial);
  recebe_no_inicial (inicial);
  aloca_sucessores (&sucessores);
#ifdef PC
  inicializa_estatistica_PC ();
#endif

  FIXE_G (inicial, 0);
  FIXE_H (inicial, calcule_h (inicial));
  acrescenta (&aberto, inicial);
  do { /* Ate achar o no meta ou fracassar */
    estatistica ();
    melhor_no = acha_melhor_no (&aberto);
    if (melhor_no != NULL) {
      acrescenta (&fechado, melhor_no);
      nos_tratados++;   /* Estatistica dos nos */
      if (calcule_h (melhor_no) == 0)
	no_meta = melhor_no;
      else {
	byte avalia = 0,       /* Indice do sucessor que esta sendo avaliado */
	     numero_sucessores;

	zera_sucessores (&sucessores);
	atualize_nivel_arvore (melhor_no); /* Registra o nivel da arvore */
	numero_sucessores = gera_sucessores (&sucessores, melhor_no);
	for (avalia = 0; avalia < numero_sucessores; avalia++) {
	  APONTAR (sucessores [avalia], melhor_no);
	  FIXE_G (sucessores [avalia], G (melhor_no) + 1);
	  FIXE_H (sucessores [avalia], calcule_h (sucessores [avalia]));
	  if (esta_em (aberto, sucessores [avalia], &antigo)) {
	    acrescenta_sucessor (melhor_no, antigo);
	    if (G (antigo) > G (sucessores [avalia])) {
	      APONTAR (antigo, melhor_no);
	      FIXE_G (antigo, G (melhor_no) + 1);
	    }
	  }
	  else
	    if (esta_em (fechado, sucessores [avalia], &antigo)) {
	      acrescenta_sucessor (melhor_no, antigo);
	      if (G (antigo) > G (sucessores [avalia])) {
		APONTAR (antigo, melhor_no);
		FIXE_G (antigo, G (melhor_no) + 1);
		expanda_f (antigo, G (antigo));
	      }
	    }
	    else {
	      sucessor_alocado = aloca_sucessor (sucessores [avalia]);
	      acrescenta (&aberto, sucessor_alocado);
	      acrescenta_sucessor (melhor_no, sucessor_alocado);
	    }
	}
      }
    }
  } while (no_meta == NULL && /* Foi encontrada uma solucao */
	   aberto != NULL);   /* Fracasso, nao ha mais nenhum nodo */
#ifdef PC
  finaliza_estatistica_PC ();
#endif
  if (no_meta != NULL)
    mostra_caminho (no_meta);
  desaloca_nos (inicial);
}
